{***************************************************************
 *
 * Unit Name: AJBFormCaption
 * Purpose  : Specify a TWinControl to act as a Caption Bar
 * Author   : Andrew Baylis
 * Date     : 6/07/2001
 * History  : Ver 1.00
 *
 ****************************************************************}

unit AJBFormCaption;

interface

uses Classes, messages, Controls, Forms, Windows;

type

  TAJBFormCaption = class(TComponent)
  private
    FActive: Boolean;
    FForm: TWinControl;
    FHideWindowCaption: Boolean;
    FNewProc: Pointer;
    FOldProc: Pointer;
    FWinControl: TWinControl;
    procedure Activate;
    procedure DeActivate;
    procedure HookedMsgProc(var Msg: TMessage);
    procedure HookWindow;
    procedure SetActive(Value: boolean);
    procedure SetHideWindowCaption(Value: Boolean);
    procedure UnhookWindow;
  protected
    procedure Loaded; override;
    procedure ProcessCaptionBar(Hide: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Active: Boolean read FActive write SetActive;
    property HideWindowCaption: Boolean read FHideWindowCaption write 
        SetHideWindowCaption;
    property WinControl: TWinControl read FWinControl write FWinControl default nil;
  end;

procedure Register;

implementation

procedure register;
begin
  RegisterComponents('AJB', [TAJBFormCaption]);
end;

procedure TAJBFormCaption.Activate;
begin
  if not (csDesigning in ComponentState) and (Owner is TControl) then
  begin
    // If the property WinControl is empty try underlying form
    if not assigned(FWinControl) then
      FWinControl := FForm;
    HookWindow;

    ProcessCaptionBar(FHideWindowCaption);
  end;
end;

procedure TAJBFormCaption.DeActivate;
begin
  UnHookWindow;
  ProcessCaptionBar(False);
end;

procedure TAJBFormCaption.HookedMsgProc(var Msg: TMessage);
var
  Handled: boolean;
begin
  Handled := False;
  if (Msg.Msg >= WM_MOUSEMOVE) and (Msg.Msg <= WM_MBUTTONDBLCLK) and
    (FForm <> nil) then begin
    Handled := True;
    // change mouse message to NC Mouse message
    Postmessage(FForm.Handle, Msg.Msg - $0160, HTCAPTION, Msg.lParam);
    Msg.Result := 0;
  end;
  // If the event isn't handled yet call old WndProc
  if not Handled then
  begin
    // Must set Msg.Result
    Msg.Result := CallWindowProc(FOldProc, WinControl.Handle, Msg.Msg, Msg.WParam, Msg.LParam);
  end;
end;

procedure TAJBFormCaption.HookWindow;
begin
  if not (csDesigning in ComponentState) and (Owner is TControl) then
  begin
    if assigned(FWinControl) and (FOldProc = nil) then
    begin
      FOldProc := Pointer(GetWindowLong(FWinControl.Handle, GWL_WNDPROC));
      SetWindowLong(FWinControl.Handle, GWL_WNDPROC, Integer(FNewProc));
    end;
  end;
end;

procedure TAJBFormCaption.SetActive(Value: boolean);
begin
  if FActive <> Value then
  begin
    FActive := Value;
    if not (csLoading in ComponentState) then
      if FActive then
        Activate
      else
        DeActivate;
  end;
end;

procedure TAJBFormCaption.SetHideWindowCaption(Value: Boolean);
begin
  FHideWindowCaption := Value;
  if (FActive) and (not (csLoading in ComponentState)) then
    ProcessCaptionBar(Value);
end;

procedure TAJBFormCaption.UnhookWindow;
begin
  if not (csDesigning in ComponentState) and (Owner is TControl) then
  begin
    if assigned(WinControl) then
    begin
      if FOldProc <> nil then
        SetWindowLong(WinControl.Handle, GWL_WNDPROC, Integer(FOldProc));
      FOldProc := nil;
    end
  end;
end;

procedure TAJBFormCaption.Loaded;
begin
  inherited Loaded;
  if FActive then
    Activate;
end;

procedure TAJBFormCaption.ProcessCaptionBar(Hide: Boolean);
var
  winStyle: Longint;
  h1: HRGN;

begin
  if not (csDesigning in ComponentState) and (Owner is TControl) and
    (FForm <> nil) then
  begin
    winStyle := GetWindowLong(FForm.Handle, GWL_STYLE);
    if Hide then
      winStyle := winStyle and (not WS_CAPTION)
    else winStyle := winStyle or WS_CAPTION;
    
    h1 := 0;
    SetWindowLong(FForm.Handle, GWL_STYLE, winStyle);
    GetWindowRgn(FForm.Handle, h1);
    SetWindowRgn(FForm.Handle, 0, True); // forces a repaint of whole window
    SetWindowRgn(FForm.Handle, h1, True); // resizes window appropriately
  end;
end;

constructor TAJBFormCaption.Create(AOwner: TComponent);
var
  tc: TWinControl;
begin
  inherited Create(AOwner);
  FWinControl := nil;
  FActive := False;
  tc := TWinControl(AOwner);
  while (tc <> nil) and (not (tc is TForm)) do tc := tc.Parent;
  if (tc <> nil) then FForm := tc
  else FForm := nil;
  // Make FNewProc a pointer to the new WndProc
  FNewProc := MakeObjectInstance(HookedMsgProc);
end;

destructor TAJBFormCaption.Destroy;
begin
  DeActivate;
  FreeObjectInstance(FNewProc);
  inherited Destroy;
end;

end.

